try:
    import chromadb  # noqa: F401
except ImportError:
    raise ValueError(
        "The chromadb is not installed. This package (chromadbx) requires that Chroma is installed to work. "
        "Please install it with `pip install chromadb`"
    )



import typer

from chroma_ops.fts import fts_commands
from chroma_ops.wal import wal_commands
from chroma_ops.db import db_commands
from chroma_ops.hnsw import hnsw_commands
from chroma_ops.collection import collection_commands

app = typer.Typer(no_args_is_help=True, help="ChromaDB Ops Commands.")


app.add_typer(wal_commands, name="wal", help="WAL maintenance commands")


app.add_typer(db_commands, name="db", help="DB maintenance commands")

app.add_typer(hnsw_commands, name="hnsw", help="HNSW index maintenance commands")
app.add_typer(
    fts_commands, name="fts", help="Full Text Search index maintenance commands"
)

app.add_typer(
    collection_commands, name="collection", help="Collection maintenance commands"
)

if __name__ == "__main__":
    app()



from enum import Enum
import json
import sqlite3
from typing import Optional
import typer

from chroma_ops.utils import (
    SqliteMode,
    get_sqlite_connection,
    print_chroma_version,
    validate_chroma_persist_dir,
)
from rich.console import Console
from rich.table import Table


class PurgeFlag(str, Enum):
    OFF = "off"
    AUTO = "auto"


def config_wal(
    persist_dir: str, *, purge: Optional[PurgeFlag] = None, yes: Optional[bool] = False
) -> None:
    validate_chroma_persist_dir(persist_dir)
    console = Console()
    print_chroma_version(console)
    with get_sqlite_connection(persist_dir, SqliteMode.READ_WRITE) as conn:
        cursor = conn.cursor()
        try:
            cursor.execute(
                """SELECT name
            FROM sqlite_master
            WHERE type='table' AND name='embeddings_queue_config';"""
            )
            cursor.fetchone()
        except sqlite3.OperationalError:
            console.print(
                "[red]WAL config table not found. Likely unsupported version of Chroma.[/red]"
            )
            return
        try:
            current_config = cursor.execute(
                """SELECT config_json_str FROM embeddings_queue_config"""
            ).fetchone()
            current_config = json.loads(current_config[0])
            del current_config["_type"]
            table = Table(title="Current WAL config")
            table.add_column("Config key", style="cyan")
            table.add_column("Config Change", style="green")

            if purge is not None:
                if (
                    purge == PurgeFlag.OFF
                    and current_config["automatically_purge"] is False
                ):
                    console.print(
                        "[bold green]WAL config is already set to the desired state (auto purge disabled).[/bold green]"
                    )
                    return
                elif (
                    purge == PurgeFlag.AUTO
                    and current_config["automatically_purge"] is True
                ):
                    console.print(
                        "[bold green]WAL config is already set to the desired state (auto purge enabled).[/bold green]"
                    )
                    return
                table.add_row(
                    "Automatically purge (automatically_purge)",
                    f"{str(current_config['automatically_purge'])} (old) -> [red]{'True' if purge == PurgeFlag.AUTO else 'False'} (new)[/red]",
                )
            else:
                table.add_row(
                    "Automatically purge (automatically_purge)",
                    f"{str(current_config['automatically_purge'])}",
                )
            console.print(table)

            if not yes:
                if not typer.confirm(
                    "\nAre you sure you want to update the WAL config?",
                    default=False,
                    show_default=True,
                ):
                    console.print("[yellow]Rebuild cancelled by user[/yellow]")
                    return
            cursor.execute("BEGIN EXCLUSIVE")
            if purge is not None:
                if purge == PurgeFlag.OFF:
                    cursor.execute(
                        """UPDATE embeddings_queue_config
                    SET config_json_str = JSON_SET(config_json_str, '$.automatically_purge', json('false'))"""
                    )
                elif purge == PurgeFlag.AUTO:
                    cursor.execute(
                        """UPDATE embeddings_queue_config
                    SET config_json_str = JSON_SET(config_json_str, '$.automatically_purge', json('true'))"""
                    )
            conn.commit()
            console.print("[bold green]WAL config updated successfully![/bold green]")
        except sqlite3.OperationalError as e:
            console.print(f"[red]Failed to update WAL config: {e}[/red]")
            conn.rollback()


def command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
    purge: PurgeFlag = typer.Option(
        PurgeFlag.AUTO,
        "--purge",
        help="Configure the purge behaviour for the WAL. Available options are: off - inifinite WAL, auto - purge WAL after adding a batch.",
    ),
    yes: bool = typer.Option(
        False,
        "--yes",
        "-y",
        help="Skip confirmation prompt",
    ),
) -> None:
    config_wal(persist_dir, purge=purge, yes=yes)



import os
import shutil
from typing import Optional

from rich.console import Console
from rich.table import Table
import typer
from chroma_ops.utils import (
    SqliteMode,
    get_sqlite_connection,
    print_chroma_version,
    validate_chroma_persist_dir,
)


def clean(persist_dir: str, yes: Optional[bool] = False) -> None:
    validate_chroma_persist_dir(persist_dir)
    with get_sqlite_connection(persist_dir, SqliteMode.READ_ONLY) as conn:
        console = Console()
        cursor = conn.cursor()
        print_chroma_version(console)
        console.print("Cleaning up orphanated segment dirs...")
        query = "SELECT id FROM segments WHERE scope = 'VECTOR';"
        cursor.execute(query)
        results = cursor.fetchall()
        active_segments = []
        for result in results:
            active_segments.append(result[0])
        cursor.close()
        conn.commit()
        # list dirs in persist_dir
        dirs_to_delete = []
        console.print()
        table = Table(title="Orphanated HNSW segment dirs")
        table.add_column("Segment ID", style="cyan")
        table.add_column("Path", style="green")
        for dir in os.listdir(persist_dir):
            if (
                os.path.isdir(os.path.join(persist_dir, dir))
                and dir not in active_segments
                and os.path.exists(os.path.join(persist_dir, dir, "header.bin"))
            ):
                dirs_to_delete.append(dir)
                table.add_row(dir, os.path.join(persist_dir, dir))
        if len(dirs_to_delete) == 0:
            console.print("[green]No orphanated segment dirs found[/green]")
            return
        console.print(table)
        if not yes:
            if not typer.confirm(
                "\nAre you sure you want to delete these segment dirs?",
                default=False,
                show_default=True,
            ):
                console.print("[yellow]Deletion cancelled by user[/yellow]")
                return
        for dir in dirs_to_delete:
            console.print(f"Deleting orphanated segment dir: {dir}")
            shutil.rmtree(os.path.join(persist_dir, dir))
        console.print("[green]Done[/green]")


def command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
    yes: bool = typer.Option(
        False,
        "--yes",
        "-y",
        help="Skip confirmation prompt",
    ),
) -> None:
    clean(persist_dir, yes)



#!/usr/bin/env python3
import os
from typing import Optional, Sequence
import typer
from chromadb import __version__ as chroma_version
from chroma_ops.utils import (
    SqliteMode,
    get_sqlite_connection,
    print_chroma_version,
    validate_chroma_persist_dir,
    get_hnsw_index_ids,
    get_dir_size,
    PersistentData,
    decode_seq_id,
)
from chroma_ops.constants import DEFAULT_TENANT_ID, DEFAULT_TOPIC_NAMESPACE
from rich.console import Console


def clean_wal(
    persist_dir: str,
    skip_collection_names: Optional[Sequence[str]] = None,
    tenant: Optional[str] = DEFAULT_TENANT_ID,
    topic_namespace: Optional[str] = DEFAULT_TOPIC_NAMESPACE,
    yes: Optional[bool] = False,
) -> None:
    validate_chroma_persist_dir(persist_dir)
    console = Console()
    print_chroma_version(console)
    console.print(f"[green]Size before: {get_dir_size(persist_dir)}[/green]")
    with get_sqlite_connection(persist_dir, SqliteMode.READ_WRITE) as conn:
        cursor = conn.cursor()
        if int(chroma_version.split(".")[1]) <= 4:
            collection_name_index = 4
            query = "SELECT s.id as 'segment',s.topic as 'topic', c.id as 'collection' , c.dimension as 'dimension', c.name FROM segments s LEFT JOIN collections c ON s.collection = c.id WHERE s.scope = 'VECTOR';"
        else:
            collection_name_index = 3
            query = "SELECT s.id as 'segment', c.id as 'collection' , c.dimension as 'dimension', c.name FROM segments s LEFT JOIN collections c ON s.collection = c.id WHERE s.scope = 'VECTOR';"

        cursor.execute(query)

        results = cursor.fetchall()
        wal_cleanup_queries = []
        if len(results) == 0:
            console.print("[green]No WAL entries found. Nothing to clean up.[/green]")
            return
        if not yes:
            if not typer.confirm(
                f"\nAre you sure you want to clean up the WAL in {persist_dir}? This action will delete all WAL entries that are not committed to the HNSW index.",
                default=False,
                show_default=True,
            ):
                console.print("[yellow]WAL cleanup cancelled by user[/yellow]")
                return
        for row in results:
            if (
                skip_collection_names
                and row[collection_name_index] in skip_collection_names
            ):
                continue
            if int(chroma_version.split(".")[1]) <= 4:
                segment_id = row[0]
                topic = row[1]
                collection_id = row[2]
            else:
                segment_id = row[0]
                collection_id = row[1]
                topic = f"persistent://{tenant}/{topic_namespace}/{collection_id}"
            metadata_pickle = os.path.join(
                persist_dir, segment_id, "index_metadata.pickle"
            )
            if os.path.exists(metadata_pickle):
                metadata = PersistentData.load_from_file(metadata_pickle)
                if hasattr(metadata, "max_seq_id"):
                    max_seq_id = metadata.max_seq_id
                else:
                    max_seq_id_query_hnsw_057 = (
                        "SELECT seq_id FROM max_seq_id WHERE segment_id = ?"
                    )
                    cursor.execute(max_seq_id_query_hnsw_057, [row[0]])
                    results = cursor.fetchall()
                    max_seq_id = decode_seq_id(results[0][0]) if len(results) > 0 else 0
                wal_cleanup_queries.append(
                    f"DELETE FROM embeddings_queue WHERE seq_id < {max_seq_id} AND topic='{topic}';"
                )
            else:
                # TODO this way of getting the hnsw space might be wrong going forward with 0.6.x
                hnsw_space = cursor.execute(
                    "select str_value from collection_metadata where collection_id=? and key='hnsw:space'",
                    (collection_id,),
                ).fetchone()
                if not hnsw_space:
                    continue
                hnsw_space = "l2" if hnsw_space is None else hnsw_space[0]
                list_of_ids = get_hnsw_index_ids(
                    f"{os.path.join(persist_dir, segment_id)}", hnsw_space, row[3]
                )
                batch_size = 100
                for batch in range(0, len(list_of_ids), batch_size):
                    wal_cleanup_queries.append(
                        f"DELETE FROM embeddings_queue WHERE seq_id IN ({','.join([str(i) for i in list_of_ids[batch:batch + batch_size]])});"
                    )
        if len(wal_cleanup_queries) > 0:
            console.print("[green]Cleaning up WAL[/green]")
            wal_cleanup_queries.insert(
                0, "BEGIN EXCLUSIVE;"
            )  # locking the DB exclusively to prevent other processes from accessing it
            wal_cleanup_queries.append("COMMIT;")
            wal_cleanup_queries.append("VACUUM;")
            cursor.executescript("\n".join(wal_cleanup_queries))
        # Close the cursor and connection
        cursor.close()
    console.print(
        f"[green]WAL cleaned up. Size after: {get_dir_size(persist_dir)}[/green]"
    )


def command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
    skip_collection_names: str = typer.Option(
        None,
        "--skip-collection-names",
        "-s",
        help="Comma separated list of collection names to skip",
    ),
    yes: Optional[bool] = typer.Option(
        False, "--yes", "-y", help="Skip confirmation prompt"
    ),
) -> None:
    clean_wal(persist_dir, skip_collection_names=skip_collection_names, yes=yes)



from typing import Optional

import chromadb
import typer

from chroma_ops.constants import DEFAULT_TOKENIZER
from chroma_ops.utils import (
    SqliteMode,
    get_sqlite_connection,
    print_chroma_version,
    validate_chroma_persist_dir,
    read_script,
)

from rich.console import Console


fts_commands = typer.Typer(no_args_is_help=True)


def validate_tokenizer(tokenizer: str) -> None:
    valid_tokenizers = ["trigram", "unicode61", "ascii", "porter"]
    if (
        not tokenizer.startswith("trigram")
        and not tokenizer.startswith("unicode61")
        and not tokenizer.startswith("ascii")
        and not tokenizer.startswith("porter")
    ):
        raise ValueError(
            f"Invalid tokenizer. Must be one of: {', '.join(valid_tokenizers)}. See https://www.sqlite.org/fts5.html#tokenizers"
        )


def rebuild_fts(
    persist_dir: str,
    tokenizer: str = DEFAULT_TOKENIZER,
    yes: Optional[bool] = False,
) -> None:
    validate_chroma_persist_dir(persist_dir)
    validate_tokenizer(tokenizer)
    console = Console()
    print_chroma_version(console)
    if not yes:
        if not typer.confirm(
            f"\nAre you sure you want to rebuild the FTS index in {persist_dir}? This action will drop the existing FTS index and create a new one.",
            default=False,
            show_default=True,
        ):
            console.print("[yellow]Rebuild FTS cancelled by user[/yellow]")
            return
    with get_sqlite_connection(persist_dir, SqliteMode.READ_WRITE) as conn:
        cursor = conn.cursor()
        script = read_script("scripts/drop_fts.sql")
        script = script.replace("__TOKENIZER__", tokenizer)
        cursor.executescript(script)
        cursor.close()
        console.print("Rebuilt FTS. Will try to start your Chroma now.")
        console.print(
            "NOTE: Depending on the size of your documents in Chroma it may take a while for Chroma to start up again.",
        )
        try:
            chromadb.PersistentClient(path=persist_dir)
            console.print("[green]Chroma started successfully. FTS rebuilt.[/green]")
        except Exception as e:
            console.print(f"[red]Chroma failed to start. Error: {repr(e)}[/red]")
            raise e


def rebuild_command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
    tokenizer: str = typer.Option(
        DEFAULT_TOKENIZER,
        "--tokenizer",
        "-t",
        help="The tokenizer to use for the FTS index. Supported values: 'trigram', 'unicode61', 'ascii', 'porter'. See https://www.sqlite.org/fts5.html#tokenizers",
    ),
    yes: Optional[bool] = typer.Option(
        False, "--yes", "-y", help="Skip confirmation prompt"
    ),
) -> None:
    rebuild_fts(persist_dir, tokenizer, yes=yes)


fts_commands.command(
    name="rebuild", help="Rebuilds Full Text Search index.", no_args_is_help=True
)(rebuild_command)



import hashlib
import os
from pathlib import Path
import sys
from typing import Optional
import zlib
import typer
from chromadb import __version__ as chroma_version
from chroma_ops.constants import DEFAULT_TENANT_ID, DEFAULT_TOPIC_NAMESPACE
from chroma_ops.utils import (
    SqliteMode,
    get_sqlite_connection,
    get_sqlite_snapshot_connection,
    print_chroma_version,
    read_script,
    validate_chroma_persist_dir,
)
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn
from rich.table import Table


def _copy_collection_to_snapshot_db(
    persist_dir: str, collection: str, output_file: Path, yes: Optional[bool] = False
) -> None:
    console = Console()
    with get_sqlite_connection(persist_dir, SqliteMode.READ_WRITE) as conn:
        collection_id = conn.execute(
            "SELECT id FROM collections WHERE name = ?",
            (collection,),
        ).fetchone()
        if not collection_id:
            raise ValueError(f"Collection {collection} not found")
        collection_id = collection_id[0]
        vector_segment_id = conn.execute(
            "SELECT id FROM segments WHERE scope = 'VECTOR' AND collection = ?",
            (collection_id,),
        ).fetchone()
        if not vector_segment_id:
            raise ValueError(f"Vector segment for collection {collection} not found")
        vector_segment_id = vector_segment_id[0]
        metadata_segment_id = conn.execute(
            "SELECT id FROM segments WHERE scope = 'METADATA' AND collection = ?",
            (collection_id,),
        ).fetchone()
        if not metadata_segment_id:
            raise ValueError(f"Metadata segment for collection {collection} not found")
        metadata_segment_id = metadata_segment_id[0]
        segment_dir = os.path.join(persist_dir, vector_segment_id)
        topic = f"persistent://{DEFAULT_TENANT_ID}/{DEFAULT_TOPIC_NAMESPACE}/{collection_id}"
        table = Table(title="Copying collection to snapshot database...")
        table.add_column("Table", style="cyan")
        table.add_column("Count", style="magenta")
        console.print(
            f"Copying collection [red]{collection}[/red] to snapshot database..."
        )
        embeddings_queue_count = conn.execute(
            "SELECT COUNT(*) FROM main.embeddings_queue WHERE topic = ?",
            (topic,),
        ).fetchone()
        table.add_row("Embeddings Queue", f"{embeddings_queue_count[0]:,}")
        max_seq_id_count = conn.execute(
            "SELECT COUNT(*) FROM main.max_seq_id WHERE segment_id IN (?, ?)",
            (vector_segment_id, metadata_segment_id),
        ).fetchone()
        table.add_row("Max Seq ID", f"{max_seq_id_count[0]:,}")
        embeddings_count = conn.execute(
            "SELECT COUNT(*) FROM main.embeddings WHERE segment_id = ?",
            (metadata_segment_id,),
        ).fetchone()
        table.add_row("Embeddings", f"{embeddings_count[0]:,}")
        embedding_metadata_count = conn.execute(
            "SELECT COUNT(*) FROM main.embedding_metadata WHERE id IN (SELECT id FROM main.embeddings WHERE segment_id = ?)",
            (metadata_segment_id,),
        ).fetchone()
        table.add_row("Embedding Metadata", f"{embedding_metadata_count[0]:,}")
        segments_count = conn.execute(
            "SELECT COUNT(*) FROM main.segments WHERE collection = ?",
            (collection_id,),
        ).fetchone()
        table.add_row("Segments", f"{segments_count[0]:,}")
        segment_metadata_count = conn.execute(
            "SELECT COUNT(*) FROM main.segment_metadata WHERE segment_id IN (SELECT id FROM main.segments WHERE collection = ?)",
            (collection_id,),
        ).fetchone()
        table.add_row("Segment Metadata", f"{segment_metadata_count[0]:,}")
        collections_count = conn.execute(
            "SELECT COUNT(*) FROM main.collections WHERE id = ?",
            (collection_id,),
        ).fetchone()
        table.add_row("Collections", f"{collections_count[0]:,}")
        collection_metadata_count = conn.execute(
            "SELECT COUNT(*) FROM main.collection_metadata WHERE collection_id = ?",
            (collection_id,),
        ).fetchone()
        table.add_row("Collection Metadata", f"{collection_metadata_count[0]:,}")
        table.add_row("HNSW Segment Data Files", f"{len(os.listdir(segment_dir)):,}")
        console.print(table)
        if not yes:
            if not typer.confirm(
                "\nAre you sure you want to copy this collection to the snapshot database?",
                default=False,
                show_default=True,
            ):
                console.print("[yellow]Copy cancelled by user[/yellow]")
                return
        conn.execute("BEGIN EXCLUSIVE")
        try:
            with Progress(
                SpinnerColumn(
                    finished_text="[bold green]:heavy_check_mark:[/bold green]"
                ),
                TextColumn("[progress.description]{task.description}"),
                *[
                    BarColumn(),
                    TextColumn("{task.percentage:>3.0f}%"),
                ],  # Add these columns
                transient=True,
            ) as progress:
                conn.execute(
                    "ATTACH DATABASE ? AS snapshot",
                    (output_file.absolute().as_posix(),),
                )
                # copy the collection to the snapshot db
                task = progress.add_task("Copying embedings_queue...", total=0)
                conn.execute(
                    "INSERT INTO snapshot.embeddings_queue SELECT * FROM main.embeddings_queue WHERE topic = ?",
                    (topic,),
                )
                progress.update(task, advance=1)
                task = progress.add_task("Copying max_seq_id...", total=0)
                conn.execute(
                    "INSERT INTO snapshot.max_seq_id SELECT * FROM main.max_seq_id WHERE segment_id IN (?, ?)",
                    (vector_segment_id, metadata_segment_id),
                )
                progress.update(task, advance=1)
                task = progress.add_task("Copying embeddings...", total=0)
                conn.execute(
                    "INSERT INTO snapshot.embeddings SELECT * FROM main.embeddings WHERE segment_id = ? ",
                    (metadata_segment_id,),
                )
                progress.update(task, advance=1)
                task = progress.add_task("Copying embedding_metadata...", total=0)
                conn.execute(
                    "INSERT INTO snapshot.embedding_metadata SELECT * FROM main.embedding_metadata WHERE id IN (SELECT id FROM main.embeddings WHERE segment_id = ?)",
                    (metadata_segment_id,),
                )
                progress.update(task, advance=1)
                task = progress.add_task("Copying segments...", total=0)
                conn.execute(
                    "INSERT INTO snapshot.segments SELECT * FROM main.segments WHERE collection = ?",
                    (collection_id,),
                )
                progress.update(task, advance=1)
                task = progress.add_task("Copying segment_metadata...", total=0)
                conn.execute(
                    "INSERT INTO snapshot.segment_metadata SELECT * FROM main.segment_metadata WHERE segment_id IN (SELECT id FROM main.segments WHERE collection = ?)",
                    (collection_id,),
                )
                progress.update(task, advance=1)
                task = progress.add_task("Copying collections...", total=0)
                conn.execute(
                    """
                    INSERT INTO snapshot.collections (id, name, dimension, database_id, database_name, tenant_id, config_json_str)
                    SELECT
                        src.id,
                        src.name,
                        src.dimension,
                        src.database_id,
                        db.name AS database_name,
                        t.id AS tenant_id,
                        src.config_json_str
                    FROM
                        main.collections AS src
                    JOIN
                        main.databases AS db ON src.database_id = db.id
                    JOIN
                        tenants AS t ON db.tenant_id = t.id
                    WHERE
                        src.id = ?;
                    """,
                    (collection_id,),
                )
                progress.update(task, advance=1)
                task = progress.add_task("Copying collection_metadata...", total=0)
                conn.execute(
                    "INSERT INTO snapshot.collection_metadata SELECT * FROM main.collection_metadata WHERE collection_id = ?",
                    (collection_id,),
                )
                progress.update(task, advance=1)

                task = progress.add_task(
                    "Copying hnsw_segment_data...", total=len(os.listdir(segment_dir))
                )
                for filename in os.listdir(segment_dir):
                    filepath = os.path.join(segment_dir, filename)
                    if os.path.isfile(filepath):
                        with open(filepath, "rb") as file:
                            binary_data = file.read()
                        compressed_data = zlib.compress(binary_data)
                        sha256 = hashlib.sha256(binary_data).hexdigest()
                        # Insert the compressed data into the database
                        conn.execute(
                            "INSERT INTO snapshot.hnsw_segment_data (segment_id, filename, data, sha256) VALUES (?, ?, ?, ?)",
                            (vector_segment_id, filename, compressed_data, sha256),
                        )
                        progress.update(task, advance=1)
            conn.commit()
            console.print(
                f"[green]Collection [red]{collection}[/red] copied to snapshot database in [red]{output_file.absolute().as_posix()}[/red][/green]"
            )
        except Exception as e:
            conn.rollback()
            raise e


def collection_snapshot(
    persist_dir: str,
    collection: str,
    output_file: Path,
    yes: Optional[bool] = False,
) -> None:
    console = Console()
    if tuple(int(part) for part in chroma_version.split(".")) < (0, 6, 0):
        console.print(
            "Collection snapshot is not supported for this version of ChromaDB",
            file=sys.stderr,
        )
        sys.exit(1)
    validate_chroma_persist_dir(persist_dir)
    print_chroma_version(console)
    if output_file.exists():
        if not yes:
            if not typer.confirm(
                f"\nAre you sure you want to overwrite {output_file.absolute().as_posix()} file?",
                default=False,
                show_default=True,
            ):
                console.print("[yellow]Snapshot cancelled by user[/yellow]")
                return
        os.remove(output_file.absolute().as_posix())
    os.makedirs(output_file.parent, exist_ok=True)
    console = Console()
    with get_sqlite_snapshot_connection(
        output_file.absolute().as_posix(), mode=SqliteMode.READ_WRITE
    ) as conn:
        console.print("Bootstrapping snapshot database...")
        script = read_script("scripts/snapshot.sql")
        conn.executescript(script)
        conn.commit()
        console.print(
            f"[green]Snapshot database bootstrapped in [red]{output_file.absolute().as_posix()}[/red][/green]"
        )
    _copy_collection_to_snapshot_db(persist_dir, collection, output_file, yes=yes)


def command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
    collection: str = typer.Option(
        ..., "--collection", "-c", help="The collection to snapshot"
    ),
    output_file: Path = typer.Option(..., "--output", "-o", help="The output file"),
    yes: Optional[bool] = typer.Option(
        False, "--yes", "-y", help="Skip confirmation prompt"
    ),
) -> None:
    collection_snapshot(persist_dir, collection, output_file, yes=yes)



#!/usr/bin/env python3
import os
from typing import Sequence, Optional

import chromadb
import typer
from rich.console import Console
from chromadb.segment.impl.vector.local_persistent_hnsw import PersistentData
from chromadb.segment import VectorReader
from chromadb.types import Operation
from chromadb import __version__ as chroma_version
from chroma_ops.constants import DEFAULT_TENANT_ID, DEFAULT_TOPIC_NAMESPACE
from chroma_ops.utils import (
    SqliteMode,
    get_sqlite_connection,
    print_chroma_version,
    validate_chroma_persist_dir,
)
from rich.table import Table


def commit_wal(
    persist_dir: str,
    skip_collection_names: Optional[Sequence[str]] = None,
    yes: Optional[bool] = False,
    tenant: Optional[str] = DEFAULT_TENANT_ID,
    topic_namespace: Optional[str] = DEFAULT_TOPIC_NAMESPACE,
) -> None:
    """Note this uses internal ChromaDB APIs which may change at any moment."""
    validate_chroma_persist_dir(persist_dir)
    skip_collection_names = skip_collection_names or []
    console = Console()
    print_chroma_version(console)
    collections_to_commit = []
    vector_segments = []
    skipped_collections_table = Table(title="Skipped Collections")
    skipped_collections_table.add_column("Collection", style="cyan")
    table = Table(title="WAL Commit Summary")
    table.add_column("Collection", style="cyan")
    table.add_column("WAL Entries", style="magenta")
    with get_sqlite_connection(persist_dir, SqliteMode.READ_ONLY) as conn:
        collections = conn.execute(
            "SELECT c.name,c.id, s.id FROM collections c left join segments s on c.id=s.collection where s.scope='VECTOR'"
        ).fetchall()
        wal_topic_groups = conn.execute(
            "SELECT topic, count(*) FROM embeddings_queue group by topic"
        ).fetchall()
        if len(collections) == 0:
            console.print("[yellow]No collections found in the database.[/yellow]")
            return
        for collection in collections:
            if collection[0] not in skip_collection_names:
                topic = f"persistent://{tenant}/{topic_namespace}/{collection[1]}"
                wal_group_count = [s[1] for s in wal_topic_groups if s[0] == topic]
                table.add_row(
                    collection[0],
                    str(wal_group_count[0] if wal_group_count else 0),
                )
                if wal_group_count:
                    collections_to_commit.append(collection)
                    vector_segments.append(
                        {"collection_id": collection[1], "id": collection[2]}
                    )
            else:
                skipped_collections_table.add_row(collection[0])
    console.print(table)
    console.print(skipped_collections_table)
    if not yes:
        if not typer.confirm(
            f"\nAre you sure you want to commit the WAL in {persist_dir}? As part of the WAL commit action your database will be migrated to currently installed version {chroma_version}.",
            default=False,
            show_default=True,
        ):
            console.print("[yellow]WAL commit cancelled by user[/yellow]")
            return
    # TODO this won't work with Rust bindings, consider alternative approach
    client = chromadb.PersistentClient(path=persist_dir)

    for s in vector_segments:
        col = client._server._get_collection(
            s["collection_id"]
        )  # load the collection and apply WAL
        client._server._manager.hint_use_collection(
            s["collection_id"], Operation.ADD
        )  # Add hint to load the index into memory
        segment = client._server._manager.get_segment(
            s["collection_id"], VectorReader
        )  # Get the segment instance
        segment._apply_batch(segment._curr_batch)  # Apply the current WAL batch

        segment._persist()  # persist the index
        PersistentData.load_from_file(
            os.path.join(persist_dir, str(s["id"]), "index_metadata.pickle")
        )  # load the metadata after persisting
        console.print(
            f"Processing index for collection {col['name']} ({s['id']}) - "
            f"total vectors in index {len(segment._index.get_ids_list())}",
        )
        segment.close_persistent_index()
    console.print("[green]WAL commit completed.[/green]")


def command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
    skip_collection_names: str = typer.Option(
        None,
        "--skip-collection-names",
        "-s",
        help="Comma separated list of collection names to skip",
    ),
    yes: Optional[bool] = typer.Option(
        False, "--yes", "-y", help="Skip confirmation prompt"
    ),
) -> None:
    commit_wal(persist_dir, skip_collection_names=skip_collection_names, yes=yes)



from contextlib import contextmanager
from enum import Enum
import os
import pickle
import shutil
import sqlite3
from typing import List, cast, Optional, Dict, Union, Generator, Any

from chromadb import Collection
import chromadb
from chromadb import __version__ as chroma_version
import hnswlib
from rich.console import Console


def validate_chroma_persist_dir(persist_dir: str) -> None:
    if not os.path.exists(persist_dir):
        raise ValueError(f"Persist dir ({persist_dir}) does not exist")
    if not os.path.exists(f"{persist_dir}/chroma.sqlite3"):
        raise ValueError(
            f"{persist_dir} does not appear to be valid ChromaDB persist directory"
        )


class SqliteMode(str, Enum):
    READ_ONLY = "ro"
    READ_WRITE = "rw"


def get_sqlite_snapshot_connection(
    snapshot_file: str, mode: Optional[SqliteMode] = SqliteMode.READ_ONLY
) -> sqlite3.Connection:
    if mode is None:
        mode = SqliteMode.READ_ONLY
    with open(snapshot_file, "a"):
        os.utime(snapshot_file, None)
    conn = sqlite3.connect(f"file:{snapshot_file}?mode={mode.value}", uri=True)
    return conn


@contextmanager
def get_sqlite_connection(
    persist_dir: str, mode: SqliteMode = SqliteMode.READ_ONLY
) -> Generator[sqlite3.Connection, Any, None]:
    """Get a connection to the SQLite database. Assumes the database is in the persist directory."""
    sql_file = os.path.join(persist_dir, "chroma.sqlite3")
    if not os.path.exists(sql_file):
        raise ValueError(f"SQLite database file ({sql_file}) does not exist")
    else:
        with open(sql_file, "a"):
            os.utime(sql_file, None)

    conn = sqlite3.connect(f"file:{sql_file}?mode={mode.value}", uri=True)
    try:
        yield conn
    finally:
        conn.close()


def get_hnsw_index_ids(filename: str, space: str = "l2", dim: int = 384) -> List[int]:
    index = hnswlib.Index(space=space, dim=dim)
    index.load_index(
        filename,
        is_persistent_index=True,
        max_elements=100000,
    )
    ids = index.get_ids_list().copy()
    index.close_file_handles()
    return cast(List[int], ids)


def read_script(script: str) -> str:
    return open(
        os.path.join(os.path.dirname(os.path.realpath(__file__)), script), "r"
    ).read()


def get_dir_size(path: str) -> int:
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            if os.path.exists(fp):
                total_size += os.path.getsize(fp)
    return total_size


def get_file_size(path: str) -> int:
    # ensure it is a file
    if not os.path.isfile(path):
        raise ValueError(f"{path} is not a file")
    return os.path.getsize(path)


SeqId = int


# TODO this works for versions of Chroma using pickle persistent data.
class PersistentData:
    """Stores the data and metadata needed for a PersistentLocalHnswSegment"""

    dimensionality: Optional[int]
    total_elements_added: int
    max_seq_id: SeqId

    id_to_label: Dict[str, int]
    label_to_id: Dict[int, str]
    id_to_seq_id: Dict[str, SeqId]

    def __init__(
        self,
        dimensionality: Optional[int],
        total_elements_added: int,
        max_seq_id: int,
        id_to_label: Dict[str, int],
        label_to_id: Dict[int, str],
        id_to_seq_id: Dict[str, SeqId],
    ):
        self.dimensionality = dimensionality
        self.total_elements_added = total_elements_added
        self.max_seq_id = max_seq_id
        self.id_to_label = id_to_label
        self.label_to_id = label_to_id
        self.id_to_seq_id = id_to_seq_id

    @staticmethod
    def load_from_file(filename: str) -> "PersistentData":
        """Load persistent data from a file"""
        with open(filename, "rb") as f:
            ret = cast(PersistentData, pickle.load(f))
            return ret


def decode_seq_id(seq_id_bytes: Union[bytes, int]) -> SeqId:
    """Decode a byte array into a SeqID"""
    if isinstance(seq_id_bytes, int):
        return seq_id_bytes

    if len(seq_id_bytes) == 8:
        return int.from_bytes(seq_id_bytes, "big")
    elif len(seq_id_bytes) == 24:
        return int.from_bytes(seq_id_bytes, "big")
    else:
        raise ValueError(f"Unknown SeqID type with length {len(seq_id_bytes)}")


# https://stackoverflow.com/a/1094933
def sizeof_fmt(num: int, suffix: str = "B") -> str:
    n: float = float(num)
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(n) < 1024.0:
            return f"{n:3.1f}{unit}{suffix}"
        n /= 1024.0
    return f"{n:.1f}Yi{suffix}"


def list_collections(client: chromadb.PersistentClient) -> List[Collection]:
    if tuple(int(part) for part in chroma_version.split(".")) < (0, 6, 0):
        return client.list_collections()  # type: ignore
    else:
        collections = []
        for collection_name in client.list_collections():
            collections.append(client.get_collection(collection_name))
        return collections


class DistanceMetric(str, Enum):
    COSINE = "cosine"
    L2 = "l2"
    IP = "ip"


def get_disk_free_space(directory: str) -> int:
    """Returns free space in bytes for the device containing the directory"""
    return shutil.disk_usage(directory).free


def check_disk_space(source_dir: str, target_dir: str) -> bool:
    """
    Checks if target directory's device has enough space for source files
    Returns (has_space, message)
    """
    source_size = get_dir_size(source_dir)
    free_space = get_disk_free_space(target_dir)
    required_space = source_size * 1.1  # Require 10% extra space to be safe

    if free_space < required_space:
        return False
    return True


def print_chroma_version(console: Console) -> None:
    """Print the ChromaDB version"""
    console.print(
        f"[bold green]ChromaDB version[/bold green]: [bold blue]{chroma_version}[/bold blue]"
    )



import typer
from chroma_ops.wal_export import command as export_command
from chroma_ops.wal_commit import command as commit_command
from chroma_ops.wal_clean import command as clean_command
from chroma_ops.wal_config import command as config_command
from chroma_ops.wal_info import command as info_command

wal_commands = typer.Typer(no_args_is_help=True)

wal_commands.command(
    name="export", no_args_is_help=True, help="Exports the WAL to a jsonl file."
)(export_command)
wal_commands.command(
    name="commit", no_args_is_help=True, help="Commit WAL to HNSW lib binary index."
)(commit_command)
wal_commands.command(
    name="clean", no_args_is_help=True, help="Cleans up WAL and VACUUM the SQLite DB."
)(clean_command)
wal_commands.command(
    name="config", no_args_is_help=True, help="Configure WAL purge behavior."
)(config_command)
wal_commands.command(
    name="info", no_args_is_help=True, help="Get information about Chroma WAL."
)(info_command)
